

#include <stdio.h>

#include "Tmk.h"

/*  This code performs NITER iterations of an NxN life board.  */

#define N          64
#define NITER      50

int (*board) [N] ;
int (*neighbours) [N] ;

void dump(int) ;

void main(int argc, char **argv)
{
  int c ;
  int blockSizeMax, blockBase, blockSize ;
  int i, y ;
  int iter ;

  while ((c = getopt(argc, argv, "")) != -1) ;

  Tmk_startup(argc, argv);

  /*  Allocate shared `board'.  */

  if (Tmk_proc_id == 0) {
    board = (int (*) [N]) Tmk_malloc(N * sizeof(int [N]));

    Tmk_distribute(&board, sizeof(int (*) [N]));
  }
  Tmk_barrier(0);

  blockSizeMax = (N + Tmk_nprocs - 1) / Tmk_nprocs ;
  blockBase    = blockSizeMax * Tmk_proc_id ;

  if(blockBase + blockSizeMax > N)
    blockSize = N - blockBase ;
  else
    blockSize = blockSizeMax ;

  /*  Initialise `board'.  */

  for(i = 0 ; i < blockSize ; i++) {
    int x   = blockBase + i ;
    for(y = 0 ; y < N ; y++) {
      if(x == N / 2 || y == N / 2)
        board [x] [y] = 1 ;
      else
        board [x] [y] = 0 ;
    }
  }
  Tmk_barrier(0);

  /*  Dump initial state of board.  */

  dump(0) ;

  /*  Allocate private `neighbours'.  */

  neighbours = (int (*) [N]) Tmk_malloc(blockSize * sizeof(int [N]));

  /*  Main update loop.  */

  for(iter = 1 ; iter <= NITER ; iter++) {

    /*  Calculate a block of neighbour sums.  */

    for(i = 0 ; i < blockSize ; i++) {
      int x   = blockBase + i ;
      int x_n = (x + N - 1) % N ;
      int x_p = (x + 1) % N ;

      for(y = 0 ; y < N ; y++) {
        int y_n = (y + N - 1) % N ;
        int y_p = (y + 1) % N ;

        neighbours [i] [y] =
            board [x_n] [y_n] + board [x_n] [y] + board [x_n] [y_p] +
            board [x]   [y_n]                   + board [x]   [y_p] +
            board [x_p] [y_n] + board [x_p] [y] + board [x_p] [y_p] ;
      }
    }
    Tmk_barrier(0);
   
    /*  Update block of board values.  */

    for(i = 0 ; i < blockSize ; i++) {
      int x   = blockBase + i ;
      for(y = 0 ; y < N ; y++) {
        int neighbour = neighbours [i] [y] ;

        if(neighbour < 2 || neighbour > 3)
          board [x] [y] = 0 ;
        if(neighbour == 3)
          board [x] [y] = 1 ;
      }
    }
    Tmk_barrier(0);

    /*  Dump current state of board.  */

    dump(iter) ;
  }

  Tmk_exit(0);
}



void dump(int iter) {

  /*  Write current state of board to a `.pgm' file.  */

  if (Tmk_proc_id == 0) {
    char fname [256] ;
    FILE* f ;
    int x, y ;

    sprintf(fname, "life%2.2d.pgm", iter) ;

    f = fopen(fname, "w") ;

    fprintf(f, "P2\n%3d  %3d\n%3d\n", N, N, 1) ;

    for(x = 0 ; x < N ; x++) {
      for(y = 0 ; y < N ; y++)
        fprintf(f, "%d ", board [x] [y]) ;
      fprintf(f, "\n") ;
    }

    fclose(f) ;
  }
}

